# stream.rb, a collection of IO streams for Ruby
# Copyright (C) 2003  Jamis Buck (jgb3@email.byu.edu)
#
# stream.rb is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# stream.rb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with stream.rb; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# =========================================================================

class InputStream
  def read( n=nil ); end
  def eof?; end
  def pos; end
  def pos=( n ); end
  def rewind; end
  def close; end
  def closed?; end

  def getc
    read( 1 )
  end

  def each_byte
    while c = getc
      yield c
    end
  end

  def buffer( size=1024 )
    BufferedInputStream.new( self, size )
  end
end


class OutputStream
  def write( s, n=nil ); end
  def flush; end
  def close; end
  def closed?; end

  def putc( c )
    write( c, 1 )
    self
  end

  def puts( s="" )
    write( s + "\n" )
    self
  end

  def <<( s )
    write( s )
    self
  end
end


class BufferedInputStream < InputStream
  def initialize( source, size=1024 )
    @source = source

    @size = 1024
    fill_buffer
  end

  def read( n=nil )
    return nil if eof?

    result = ""

    to_eof = n.nil?
    n = 0 if to_eof

    while n > 0 or ( to_eof and not eof? )
      ensure_buffer

      bytes_to_read = ( to_eof ? @buffer.length : n )
      remaining = bytes_remaining

      bytes_to_read = remaining if bytes_to_read > remaining

      result << @buffer[ @buffer_pos, bytes_to_read ]
      @buffer_pos += bytes_to_read
      n -= bytes_to_read
    end

    result
  end

  def eof?
    @buffer_pos >= @buffer.length && @source.eof?
  end

  def pos
    @buffer_start + @buffer_pos
  end

  def pos=( n )
    @source.pos = n
    fill_buffer
  end

  def rewind
    @source.rewind
    fill_buffer
  end

  def close
    @source.close
  end

  def closed?
    @source.closed?
  end

  def getc
    return nil if eof?
    ensure_buffer
    c = @buffer[ @buffer_pos ]
    @buffer_pos += 1
    return c
  end

  def readline( sep=$/ )
    return nil if eof?

    eol = @buffer.index( sep, @buffer_pos )
    text = ""

    while eol.nil? and not eof?
      text << @buffer[ @buffer_pos..-1 ]
      fill_buffer
      eol = @buffer.index( sep, @buffer_pos )
    end

    eol = @buffer.length if eol.nil?
    text << @buffer[ @buffer_pos..eol ]
    @buffer_pos = eol+1

    return text
  end

  def each_line( sep=$/ )
    while line = readline( sep )
      yield line
    end
  end

  def readlines( sep=$/ )
    lines=[]

    each_line( sep ) { |line| lines.push line }

    return lines
  end

  alias :each :each_line
  alias :gets :readline

  # -- PRIVATE ------------------------------------------------

  def ensure_buffer
    fill_buffer if @buffer_pos >= @buffer.length
  end

  def fill_buffer
    @buffer_pos = 0
    @buffer_start = @source.pos
    @buffer = ( @source.read( @size ) || "" )
  end

  def bytes_remaining
    @buffer.length - @buffer_pos
  end

  private :fill_buffer
  private :bytes_remaining
end


class StringInputStream < InputStream
  def initialize( source )
    @source = source
    @pos = 0
  end

  def read( n=nil )
    return nil if eof?

    n = @source.length if n.nil?
    remaining = ( @source.length - pos )

    n = remaining if n > remaining

    text = @source[ @pos, n ]
    @pos += n

    return text
  end

  def eof?
    @pos >= @source.length
  end

  def pos
    @pos
  end

  def pos=( n )
    @pos = n
  end

  def rewind
    @pos = 0
  end

  def close
  end

  def closed?
    false
  end

  def getc
    c = @source[ @pos, 1 ]
    @pos += 1
    return c
  end
end


class StringOutputStream < OutputStream
  attr_reader :contents

  def initialize( buffer )
    @contents = buffer
  end

  def write( s, n=nil )
    @contents << ( n.nil? ? s : s[0,n] )
    self
  end

  def closed?
    false
  end
end


class FileInputStream < InputStream
  def initialize( filename )
    @file = File.new( filename, "r" )
  end

  def read( n=nil )
    @file.read( n )
  end

  def eof?
    @file.eof?
  end

  def pos
    @file.pos
  end

  def pos=( n )
    @file.pos = n
  end

  def rewind
    @file.rewind
  end

  def close
    @file.close
  end

  def closed?
    @file.closed?
  end

  def getc
    @file.getc
  end

  def each_byte( &block )
    @file.each_byte( &block )
  end
end
